%{
/*
	Parser to read the Makefile.in files.

	(c) Keith Owens <kaos@ocs.com.au> January 2002.
	Released under GPL version 2, (you know the rest).

	Input is the combined set of Makefile.in files.  Output is the converted
	global makefile in .tmp_global_makefile_in.2.
 */

#include "pp_makefile.h"
#include "pp_db.h"

#include <limits.h>

static const char *program = "pp_makefile2_parse.y";

/* Feed forward from lexer to parser */

extern char *l_makefile_in_name;
extern char *l_curr_reldir;
extern int l_line_no;

/* Feed back from parser to lexer */

int y_errors;
int y_depth;			/* how deep to look into parenthesis, anything deeper is text */
int y_saving_text, y_suppress_nl;
FILE *yyout, *y_feedback;
char **y_user_target, **y_user_dependencies;
PPDB *y_db_main;

/* Local variables */

static int src_base;
static int make_target;
static int indent;
static char **config_list;
static char **user_command, **user_flags;
static char **shipped_prefix, **shipped_generated, **shipped_generating,
    **shipped_command, **shipped_config;
static char **hostld_binary, **hostccld_binary, **hostccld_objects;

static char feedback[] = ".tmp_pp_makefile2_parse.y_to_l";

/* Ugly global state.  Look at rewriting this later.  KAO.  */

static char **saved_text;
static int num_saved_text;

static void
free_saved_text(char ***saved)
{
	char **s = *saved;
	if (s) {
		while (*s) {
			free(*s++);
		}
		free(*saved);
	}
	*saved = NULL;
	y_saving_text = 0;
	num_saved_text = 0;
}

static void
copy_saved_text(char ***save)
{
	if (num_saved_text) {
		*save = k_malloc((num_saved_text + 1) * sizeof (**save));
		memcpy(*save, saved_text,
		       (num_saved_text + 1) * sizeof (**save));
		free(saved_text);
	} else {
		*save = NULL;
	}
	saved_text = NULL;
	num_saved_text = 0;
}

static void
print_saved_text(FILE * fout, char **save, int quote)
{
	if (save) {
		while (*save) {
			while (strchr(*save, '\n')) {
				*(strchr(*save, '\n')) = ' ';
			}
			if (!quote) {
				fprintf(fout, "%s", *save);
			} else {
				char *p;
				for (p = *save; *p; ++p) {
					if (*p == '#') {
						fprintf(fout, "\\");
					}
					fprintf(fout, "%c", *p);
				}
			}
			++save;
		}
	}
}

static void
start_saving_text(void)
{
	free_saved_text(&saved_text);
	y_saving_text = 1;
}

static void
stop_saving_text(void)
{
	y_saving_text = 0;
}

static void
suppress_nl(void)
{
	y_suppress_nl = 1;
}

static int
save_text(const char *name)
{
	if (y_saving_text) {
		saved_text =
		    k_realloc(saved_text,
			      (num_saved_text + 2) * sizeof (*saved_text));
		saved_text[num_saved_text] = k_strdup(name);
		saved_text[num_saved_text + 1] = NULL;
		++num_saved_text;
		return (1);
	}
	return (0);
}

int
y_printf(const char *format, ...)
{
	static int printf_size;
	static char *printf_text;
	va_list ap;
	int ret;
	if (!printf_text) {
		printf_text = k_realloc(printf_text, printf_size += 4096);
	}
	/* Cannot check vnsprintf return code for overflow, it varies between systems.
	 * Instead just increase the buffer size until the text is seen to fit.
	 */
	while (1) {
		va_start(ap, format);
		ret = vsnprintf(printf_text, printf_size, format, ap);
		if (strlen(printf_text) >= printf_size - 1) {
			printf_text = k_realloc(printf_text, printf_size +=
						4096);
		} else {
			va_end(ap);
			break;
		}
	}
	if (!save_text(printf_text)) {
		fprintf(yyout, "%s", printf_text);
	}
	return (ret);
}

static void
do_indent(void)
{
	y_printf("%*s", 2 * indent, "");
}

static void
do_ifsel(char *config, int nl)
{
	do_indent();
	if (*config == '!') {
		y_printf("ifeq ($(subst n,,$(%s)),)", config + 1);
	} else {
		y_printf("ifneq ($(subst n,,$(%s)),)", config);
	}
	if (nl) {
		y_printf("\n");
		++indent;
	}
}

static void
do_endif(int n)
{
	if (n > 0 && indent > 0) {
		while (n-- && indent--) {
			do_indent();
			y_printf("endif\n");
		}
	}
}

static void
do_nodepend(const char *name)
{
	DB_TOKEN token;
	DB_REC_TARGET *t;
	token = lookup_db_target(y_db_main, name);
	if (DB_IS_NULL_TOKEN(token)) {
		fprintf(stderr, "%s: Illegal name '%s' in nodepend\n", program,
			name);
		++errors;
	} else {
		t = read_db_target(y_db_main, token);
		t->flags.nodepend = 1;
	}
}

static void
normal_config(const char *config)
{
	if (*config == '!') {
		yyerror("Negated CONFIG_ is not allowed here");
	}
}

void
y_normalize_path(char *path)
{
	char *p1, *p2;
	p1 = path + (*path == '/');
	while (*p1) {
		if (strncmp(p1, "/", 1) == 0) {
			strcpy(p1, p1 + 1);
		} else if (strncmp(p1, "./", 2) == 0) {
			strcpy(p1, p1 + 2);
		} else if (strcmp(p1, ".") == 0) {
			*p1 = '\0';
		} else if (strncmp(p1, "../", 3) == 0 || strcmp(p1, "..") == 0) {
			/* To strip out "/../", scan backwards to find the "/" at the start of the
			 * previous qualifier, the "/../" removes this qualifier.  "/../" at the
			 * start of the path is valid and has no effect, the parent directory of
			 * "/" is itself.
			 */
			p2 = p1 - 1;
			while (p2 > path && *--p2 != '/') ;
			strcpy(p2 + 1, *(p1 + 2) == '/' ? p1 + 3 : p1 + 2);
			p1 = p2 + 1;
		} else {
			p1 = strchr(p1, '/');
			if (p1) {
				++p1;
			} else {
				p1 = path + strlen(path);
			}
		}
	}
}

static void
startonefile(char *text)
{
	y_printf("__pp_extra_aflags_list\t:=\n");
	y_printf("__pp_extra_cflags_list\t:=\n");
	y_printf("__pp_extra_ldflags_list\t:=\n");
	y_printf("__pp_select_list\t:=\n");
	y_printf("__pp_objlink_list\t:=\n");
	y_printf("__pp_expsyms_list\t:=\n");
	y_printf("%s", text);
	free(text);
	y_depth = 1;
	indent = 0;
}

static void
endonefile(void)
{
	y_printf("\n");
	y_printf("__pp_makefile_dummy\t:= $(shell ( \\\n");
	y_printf("\t\t\t\techo 'reldir' %s && \\\n", l_curr_reldir);
	y_printf
	    ("\t\t\t\techo 'extra_aflags $(__pp_extra_aflags_list)' && \\\n");
	y_printf
	    ("\t\t\t\techo 'extra_cflags $(__pp_extra_cflags_list)' && \\\n");
	y_printf
	    ("\t\t\t\techo 'extra_ldflags $(__pp_extra_ldflags_list)' && \\\n");
	y_printf("\t\t\t\techo 'select $(__pp_select_list)' && \\\n");
	y_printf("\t\t\t\techo 'objlink $(__pp_objlink_list)' && \\\n");
	y_printf("\t\t\t\techo 'expsyms $(__pp_expsyms_list)' && \\\n");
	y_printf("\t\t\t\techo \\\n");
	y_printf("\t\t\t    ) >>.tmp_select )\n");
	y_printf("#__pp_targets\n");
	free(l_curr_reldir);
}

static void
standardize_name(char *name)
{
	if (strstr(name, "/../") || strstr(name, "//")) {
		char *p;
		for (p = name; (p = strstr(p, "//")); strcpy(p + 1, p + 2)) {
		};
		while ((p = strstr(name, "/../"))) {
			if (p == name) {
				strcpy(name + 1, name + 4);
			} else {
				char *p1;
				for (p1 = p - 1; *p1 != '/' && p1 >= name;
				     --p1) {
				};
				strcpy(p1 + 1, p + 4);
			}
		}
		for (p = name; (p = strstr(p, "//")); strcpy(p + 1, p + 2)) {
		};
	}
}


/* Add a target name to the object tree if it does not already exist.  This adds
 * generated input files to the database.
 *
 * Every target can potentially be used as a source, i.e. a generated
 * source.
 */

DB_REC_TARGET *
y_add_make_target(const char *dir, const char *target)
{
	int ld = strlen(dir);
	int lt = strlen(target);
	char name[ld + lt + 1];
	DB_TOKEN token;
	DB_REC_TARGET *t;
	DB_REC_SOURCE *s;
	memcpy(name, dir, ld);
	memcpy(name + ld, target, lt + 1);
	standardize_name(name);
	token = findnew_db_target(y_db_main, name);
	t = read_db_target(y_db_main, token);
	t->flags.makefile = 1;
	t->flags.ignore = 0;
	return (t);
}

/* Replace an old string with a new string, resizing the input string if
 * necessary.
 */

void
y_new_string(char **is, char **is_end, int *is_size,
	     char *old, char *old_end, const char *new)
{
	int new_len = strlen(new), len_diff = new_len - (old_end - old);
	if (len_diff <= 0) {
		/* Insert with left shift */
		memcpy(old, new, new_len);
		strcpy(old + new_len, old_end);
		*is_end += len_diff;
	} else {
		/* Insert with right shift, might need to resize */
		if ((*is_end - *is) + len_diff > *is_size) {
			char *is_new = k_malloc(*is_size + len_diff);
			int ptr_adjust = is_new - *is;
			memcpy(is_new, *is, *is_size);
			free(*is);
			*is = is_new;
			*is_end += ptr_adjust;
			*is_size += len_diff;
			*old += ptr_adjust;
			*old_end += ptr_adjust;
		}
		memmove(old_end + len_diff, old_end, strlen(old_end) + 1);
		memcpy(old, new, new_len);
		*is_end += len_diff;
	}
}
%}

%union {
  char *string;		/* pointer to k_strdup(yytext) */
}

%token EXPSYMS
%token SELECT SELECT_ELSEWHERE
%token LINK_SUBDIRS
%token EXTRA_AFLAGS EXTRA_AFLAGS_ALL
%token EXTRA_CFLAGS EXTRA_CFLAGS_ALL
%token EXTRA_LDFLAGS
%token OBJLINK
%token USER_COMMAND NODEPEND CREATE_OBJDIR SIDE_EFFECT SHIPPED ISSRCFILE
%token BASE_TARGET UPDATE_IF_CHANGED USES_ASM_OFFSETS
%token HOSTCC HOSTLD HOSTCCLD
%token SETUP INSTALLABLE
%token IFSEL IFNSEL IFSELMOD IFSELNMOD
%token LPAREN RPAREN 	/* '(', ')' */
%token EOL ENDONEFILE
%token <string> STARTONEFILE
%token <string> CONFIG
%token <string> TEXT

%%

files
	: file
	| files file
	;

file
	: STARTONEFILE
	  { startonefile($1); }
	  restoffile
	;

restoffile
	: lines
	  ENDONEFILE
	  { endonefile(); }
	| lines
	  error
	  ENDONEFILE
	  { endonefile(); }
	  { fprintf(stderr, "%s: Skipping to end of file\n", program); ++y_errors; }
	;

lines
	: line
	| lines line

line
	: units EOL
	| EOL
	;

units
	: unit
	| units unit
	;

unit
	: expsyms
	| select
	| select_elsewhere
	| link_subdirs
	| extra_aflags
	| extra_aflags_all
	| extra_cflags
	| extra_cflags_all
	| extra_ldflags
	| objlink
	| user_command
	| nodepend
	| create_objdir
	| side_effect
	| shipped
	| issrcfile
	| base_target
	| update_if_changed
	| uses_asm_offsets
	| hostcc
	| hostld
	| hostccld
	| setup
	| installable
	| ifsel
	| ifnsel
	| ifselmod
	| ifselnmod
	| text
	;

expsyms
	: EXPSYMS
	  { y_printf("__pp_expsyms_list += "); make_target = 0; }
	  LPAREN
	  local_object_list
	  RPAREN
	;

select
	: SELECT
	  LPAREN
	  { start_saving_text(); }
	  optional_config_list
	  {
		stop_saving_text();
		copy_saved_text(&config_list);
		start_saving_text();
		make_target = 0;
	  }
	  local_object_list
	  RPAREN
	  select_common
	  ;

  /* select_elsewhere kludge start.  Remove as soon as possible. */

select_elsewhere
	: SELECT_ELSEWHERE
	  LPAREN
	  { start_saving_text(); }
	  optional_config_list
	  {
		stop_saving_text();
		copy_saved_text(&config_list);
		start_saving_text();
		make_target = 0;
	  }
	  any_object_list
	  RPAREN
	  select_common
	  ;

any_object_list
	: any_object
	| any_object_list any_object
	;

any_object
	: TEXT
	  {
		int l = strlen($1);
		if (strchr($1, '$')) {
			y_printf("$(patsubst %%,%s%%,%s)", l_curr_reldir + make_target,
				 $1);
		} else {
			if (l < 3 || $1[l - 2] != '.' || !strchr("oais", $1[l - 1])) {
				yyerror("not an object name");
			}
			if (*($1) == '/') {
				y_printf("%s", $1);
				y_add_make_target("", $1);
			} else {
				yyerror("select_elsewhere() objects must start with '/'");
			}
		}
		free($1);
	  }
	;

  /* select_elsewhere kludge end.  Remove as soon as possible. */

  /* Common select code split off to support select_elsewhere.  Collapse back
   * into select when the select_elsewhere kludge is removed.
   */

select_common
	:
	  {
		char **s = config_list, *t, *first;
		int configs = 0, i;
		stop_saving_text();
		for (s = config_list; s && *s; ++s) {
			if (*(*s + strspn(*s, " \t\n"))) {
				++configs;
			}
		}
		first = NULL;
		for (s = config_list; s && *s; ++s) {
			if (*(first = *s + strspn(*s, " \t\n"))) {
				break;
			}
			first = NULL;
		}
		if (first) {
			for (++s; *s; ++s) {
				if (*(t = *s + strspn(*s, " \t\n"))) {
					do_ifsel(t, 1);
				}
			}
			if (*first == '!') {
				/* Negated first config can never select as module */
				do_ifsel(first, 1);
				++configs;
				do_indent();
				fprintf(yyout, "__pp_select_list += (y_ ");
			} else {
				do_indent();
				fprintf(yyout, "__pp_select_list += ($(%s)_ ", first);
			}
		} else {
			fprintf(yyout, "__pp_select_list += (y_ ");
		}
		/* saved_text contains object list */
		print_saved_text(yyout, saved_text, 0);
		fprintf(yyout, ")\n");
		do_endif(configs - 1);
		free_saved_text(&saved_text);
		free_saved_text(&config_list);
		suppress_nl();
	  }
	;

link_subdirs
	: LINK_SUBDIRS
	  LPAREN
	  { y_printf("__pp_select_list += (s_ "); }
	  link_subdir_list
	  RPAREN
	  { y_printf(")"); }
	;

extra_aflags
	: EXTRA_AFLAGS
	  LPAREN
	  { y_printf("__pp_extra_aflags_list += ("); make_target = 0; }
	  local_object
	  text_list
	  RPAREN
	  { y_printf(")"); }
	;

extra_aflags_all
	: EXTRA_AFLAGS_ALL
	  LPAREN
	  { y_printf("__pp_extra_aflags_list += (%s ", l_curr_reldir); }
	  text_list
	  RPAREN
	  { y_printf(")"); }
	;

extra_cflags
	: EXTRA_CFLAGS
	  LPAREN
	  { y_printf("__pp_extra_cflags_list += ("); make_target = 0; }
	  local_object
	  text_list
	  RPAREN
	  { y_printf(")"); }
	;

extra_cflags_all
	: EXTRA_CFLAGS_ALL
	  LPAREN
	  { y_printf("__pp_extra_cflags_list += (%s ", l_curr_reldir); }
	  text_list
	  RPAREN
	  { y_printf(")"); }
	;

extra_ldflags
	: EXTRA_LDFLAGS
	  LPAREN
	  { y_printf("__pp_extra_ldflags_list += ("); make_target = 0; }
	  local_target
	  text_list
	  RPAREN
	  { y_printf(")"); }
	;

objlink
	: OBJLINK
	  LPAREN
	  { start_saving_text(); }
	  optional_config_list
	  {
		stop_saving_text();
		copy_saved_text(&config_list);
		start_saving_text();
		make_target = 0;
	  }
	  any_target_list
	  RPAREN
	  {
		char **s = config_list, *t;
		int configs = 0, i;
		stop_saving_text();
		for (s = config_list; s && *s; ++s) {
			if (*(t = *s + strspn(*s, " \t\n"))) {
				++configs;
				do_ifsel(t, 1);
			}
		}
		do_indent();
		fprintf(yyout, "__pp_objlink_list += (");
		/* saved_text contains object list */
		print_saved_text(yyout, saved_text, 0);
		fprintf(yyout, ")\n");
		do_endif(configs);
		free_saved_text(&saved_text);
		free_saved_text(&config_list);
	  }
	;

user_command
	: USER_COMMAND
	  LPAREN
	    { start_saving_text(); make_target = 1; }
	    any_target
	    { copy_saved_text(&y_user_target); y_depth = 2; /* three fields nested in '()' */ }
	    LPAREN
	      { start_saving_text(); }
	      optional_text_list
	      { copy_saved_text(&y_user_dependencies); }
	    RPAREN
	    LPAREN
	      { start_saving_text(); }
	      text_list
	      { copy_saved_text(&user_command); }
	    RPAREN
	    LPAREN
	      { start_saving_text(); }
	      optional_text_list
	      { copy_saved_text(&user_flags); }
	    RPAREN
	  RPAREN
	  {
		char **cmd;
		stop_saving_text();
		fprintf(yyout, "KCN_%s := ", *y_user_target);
		/* Convert $(AS) to $(AS_real).  Ditto LD, CC, CPP, AR, STRIP,
		 * OBJCOPY, HOSTAS, HOSTCC, HOSTLD, HOSTAR.  See checks for not
		 * writable in the top level Makefile.
		 */
		for (cmd = user_command; *cmd; ++cmd) {
			if (strncmp(*cmd, "$(", 2) == 0 &&
			    (strcmp(*cmd, "$(AS)") == 0 ||
			     strcmp(*cmd, "$(LD)") == 0 ||
			     strcmp(*cmd, "$(CC)") == 0 ||
			     strcmp(*cmd, "$(CPP)") == 0 ||
			     strcmp(*cmd, "$(AR)") == 0 ||
			     strcmp(*cmd, "$(STRIP)") == 0 ||
			     strcmp(*cmd, "$(OBJCOPY)") == 0 ||
			     strcmp(*cmd, "$(HOSTAS)") == 0 ||
			     strcmp(*cmd, "$(HOSTCC)") == 0 ||
			     strcmp(*cmd, "$(HOSTLD)") == 0 ||
			     strcmp(*cmd, "$(HOSTAR)") == 0)) {
				int l = strlen(*cmd);
				char *new_cmd = k_malloc(l + 6);
				memcpy(new_cmd, *cmd, l - 1);
				memcpy(new_cmd + l - 1, "_real)", 7);
				free(*cmd);
				*cmd = new_cmd;
			}
		}
		print_saved_text(yyout, user_command, 1);
		fprintf(yyout, " ");
		print_saved_text(yyout, user_flags, 1);
		fprintf(yyout, "\n\n");
		{
			DB_TOKEN token;
			int l = strlen(*y_user_target);
			char name[l + 2];
			*name = '/';
			memcpy(name + 1, *y_user_target, l + 1);
			token = lookup_db_target(y_db_main, name);
			assert(!(DB_IS_NULL_TOKEN(token)));
			if (!(read_db_target(y_db_main, token)->flags.setup)) {
				fprintf(yyout, "%s: .tmp_setup\n", *y_user_target);
			}
		}
		fprintf(yyout, "%s: ", *y_user_target);
		print_saved_text(yyout, y_user_dependencies, 0);
		fprintf(yyout,
			"\n\t$(KBUILD_QUIET)scripts/pp_makefile5 $(PP_MAKEFILE5_FLAGS) --type=USER --target=%s --command='",
			*y_user_target);
		print_saved_text(yyout, user_command, 0);
		fprintf(yyout, "' --flags='");
		print_saved_text(yyout, user_flags, 0);
		fprintf(yyout, "'\n\nCLEAN += %s\n", *y_user_target);
		free_saved_text(&y_user_target);
		free_saved_text(&y_user_dependencies);
		free_saved_text(&user_command);
		free_saved_text(&user_flags);
	  }
	;

nodepend
	: NODEPEND
	  LPAREN
	  { start_saving_text(); make_target = 0; }
	  any_target_list
	  RPAREN
	  {
		char **s = saved_text, *t;
		while (s && *s) {
			t = *s + strspn(*s, " \t\n");
			if (*t) {
				do_nodepend(t);
			}
			++s;
		}
		free_saved_text(&saved_text);
	  }
	;

create_objdir
	: CREATE_OBJDIR
	  LPAREN
	  { start_saving_text(); make_target = 0; }
	  optional_config_list
	  RPAREN
	  {
		char **s = saved_text, *t;
		int configs = 0;
		stop_saving_text();
		while (s && *s) {
			t = *s + strspn(*s, " \t\n");
			if (*t) {
				++configs;
				do_ifsel(t, 1);
			}
			++s;
		}
		do_indent();
		fprintf(yyout, "__pp_select_list += (c_ %s)\n", l_curr_reldir);
		do_endif(configs);
		free_saved_text(&saved_text);
	  }
	;

side_effect
	: SIDE_EFFECT
	  LPAREN
	  { start_saving_text(); make_target = 0; }
	  any_target
	  any_target
	  RPAREN
	  {
		char **s, *target1 = NULL, *target2 = NULL;
		for (s = saved_text; s && *s; ++s) {
			if (!isspace(**s)) {
				target1 = *s;
				++s;
				break;
			}
		}
		for (; s && *s; ++s) {
			if (!isspace(**s)) {
				target2 = *s;
				++s;
				break;
			}
		}
		fprintf(yyout, "%s: %s\n\t@/bin/true\n\n", target1 + 1, target2 + 1);
		fprintf(yyout, "# nodepend(%s)\n", target1);
		do_nodepend(target1);
		fprintf(yyout, "CLEAN\t+= %s\n", target1 + 1);
		free_saved_text(&saved_text);
	  }
	;

shipped
	: SHIPPED
	  LPAREN
	    { start_saving_text(); }
	    text
	    { copy_saved_text(&shipped_prefix); y_depth = 2; /* four fields nested in '()' */ }
	    LPAREN
	      { start_saving_text(); }
	      text_list
	      { copy_saved_text(&shipped_generated); }
	    RPAREN
	    LPAREN
	      { start_saving_text(); }
	      text_list
	      { copy_saved_text(&shipped_generating); }
	    RPAREN
	    LPAREN
	      { start_saving_text(); }
	      text_list
	      { copy_saved_text(&shipped_command); }
	    RPAREN
	    LPAREN
	      { start_saving_text(); }
	      optional_config_list
	      { copy_saved_text(&shipped_config); }
	    RPAREN
	  RPAREN
	  {
		char **text;
		int generated = 0;
		stop_saving_text();
		if (!(y_feedback = fopen(feedback, "w+"))) {
			fprintf(stderr, "%s: Cannot create %s: %m\n", program,
				feedback);
			++errors;
			y_feedback = yyout;
		}
		fprintf(y_feedback, "%s_files :=", *shipped_prefix);
		for (text = shipped_generating; *text; ++text) {
			if (!isspace(**text)) {
				fprintf(y_feedback, " $(srcfile %s)", *text);
			}
		}
		for (text = shipped_generated; *text; ++text) {
			if (!isspace(**text)) {
				fprintf(y_feedback, " $(srcfile %s_shipped)", *text);
				++generated;
			}
		}
		/* It was much easier to pipe the sed output directly into md5sum
		 * but not all versions of md5sum support input from stdin, have to
		 * use a temporary file.
		 *
		 * Strip out RCS identifiers, have to use ABD-Z because the kernel has
		 * some places that contain $CONFIG_... .... $CONFIG which would match
		 * $[A-Z][^$]*\$.
		 */
		fprintf(y_feedback, "\n");
		fprintf(y_feedback,
			"$(objfile %s_sum.d): $(srcfile %s_sum) $(%s_files)\n",
			*shipped_prefix, *shipped_prefix, *shipped_prefix);
		fprintf(y_feedback, "\t$(KBUILD_QUIET)(set -e; \\\n");
		fprintf(y_feedback, "\t  mkdir -p $(objdir); \\\n");
		fprintf(y_feedback,
			"\t  sed -e 's:$$[ABD-Z][^$$]*\\$$::g' $(%s_files) > $(objfile .tmp_%s_files); \\\n",
			*shipped_prefix, *shipped_prefix);
		fprintf(y_feedback,
			"\t  ($(MD5SUM) -c $(srcfile %s_sum) >/dev/null 2>&1 && echo S OK || echo F) > $@; \\\n",
			*shipped_prefix);
		fprintf(y_feedback, "\t  rm $(objfile .tmp_%s_files); \\\n",
			*shipped_prefix);
		fprintf(y_feedback, "\t)\n");
		fprintf(y_feedback, "\n");
		for (text = shipped_generated; *text; ++text) {
			if (!isspace(**text)) {
				fprintf(y_feedback, "user_command(%s\n", *text);
				break;
			}
		}
		fprintf(y_feedback, "\t($(objfile %s_sum.d))\n", *shipped_prefix);
		fprintf(y_feedback, "\t([ \"`cat $<`\" = \"S OK\" ] &&\n");
		fprintf(y_feedback, "\t  (\n");
		fprintf(y_feedback, "\t    set -e;\n");
		for (text = shipped_generated; *text; ++text) {
			if (!isspace(**text)) {
				fprintf(y_feedback,
					"\t    cp -f $(srcfile %s_shipped) $(objfile %s);\n",
					*text, *text);
			}
		}
		fprintf(y_feedback, "\t  ) ||\n");
		fprintf(y_feedback, "\t  (");
		print_saved_text(y_feedback, shipped_command, 0);
		fprintf(y_feedback, "))\n");
		fprintf(y_feedback, "\t()\n");
		fprintf(y_feedback, "\t)\n");
		fprintf(y_feedback, "\n");
		fprintf(y_feedback,
			"# Exception: this writes to the source tree.  The target must be\n");
		fprintf(y_feedback, "# explicitly requested.\n");
		fprintf(y_feedback, ".PHONY: %s_sum_shipped\n", *shipped_prefix);
		fprintf(y_feedback, "%s_sum_shipped:", *shipped_prefix);
		for (text = shipped_generated; *text; ++text) {
			if (!isspace(**text)) {
				fprintf(y_feedback, " $(objfile %s)", *text);
			}
		}
		fprintf(y_feedback, "\n");
		for (text = shipped_generated; *text; ++text) {
			if (!isspace(**text)) {
				fprintf(y_feedback,
					"\t@cmp -s $(objfile %s) $(srcfile %s_shipped) || \\\n",
					*text, *text);
				fprintf(y_feedback,
					"\t     mv -f $(objfile %s) $(srcfile %s_shipped)\n",
					*text, *text);
			}
		}
		fprintf(y_feedback,
			"\t@sed -e 's:$$[ABD-Z][^$$]*\\$$::g' $(%s_files) > $(objfile .tmp_%s_files); \\\n",
			*shipped_prefix, *shipped_prefix);
		fprintf(y_feedback,
			"\t  $(MD5SUM) $(objfile .tmp_%s_files) > $(objfile .tmp_$(notdir $@)) && \\\n",
			*shipped_prefix);
		fprintf(y_feedback,
			"\t  (cmp -s $(objfile .tmp_$(notdir $@)) $(srcfile %s_sum) || \\\n",
			*shipped_prefix);
		fprintf(y_feedback, "\t    (echo Updating checksum for %s; \\\n",
			*shipped_prefix);
		fprintf(y_feedback,
			"\t    mv -f $(objfile .tmp_$(notdir $@)) $(srcfile %s_sum)))\n",
			*shipped_prefix);
		fprintf(y_feedback,
			"\t@rm -f $(objfile %s_sum.d) $(objfile .tmp_$(notdir $@) $(objfile .tmp_%s_files))\n",
			*shipped_prefix, *shipped_prefix);
		fprintf(y_feedback, "\n");
		if (shipped_config) {
			for (text = shipped_config; *text; ++text) {
				if (!isspace(**text)) {
					/* Negated config here makes little sense but it is easier to
					 * allow it than flagging an error condition at the end of the
					 * shipped() command.
					 */
					if (**text == '!') {
						fprintf(y_feedback, "ifnsel(%s)\n",
							*text + 1);
					} else {
						fprintf(y_feedback, "ifsel(%s)\n",
							*text);
					}
					fprintf(y_feedback,
						"  all_sum_shipped: %s_sum_shipped\n",
						*shipped_prefix);
					fprintf(y_feedback, "endif\n");
				}
			}
			fprintf(y_feedback, "\n");
		}
		if (generated > 1) {
			const char *prev;
			for (text = shipped_generated;; ++text) {
				if (!isspace(**text)) {
					prev = *text;
					break;
				}
			}
			for (++text; *text; ++text) {
				if (!isspace(**text)) {
					fprintf(y_feedback, "side_effect(%s %s)\n",
						*text, prev);
					prev = *text;
				}
			}
		}
		free_saved_text(&shipped_prefix);
		free_saved_text(&shipped_generated);
		free_saved_text(&shipped_generating);
		free_saved_text(&shipped_command);
		free_saved_text(&shipped_config);
		if (y_feedback == yyout) {
			y_feedback = NULL;
		} else {
			/* y_feedback is read and closed in the lexer */
			fseek(y_feedback, 0, SEEK_SET);
		}
	  }
	;

issrcfile
	: ISSRCFILE
	  LPAREN
	  { start_saving_text(); make_target = 0; }
	  any_target_list
	  RPAREN
	  {
		char **s = saved_text, *name;
		while (s && *s) {
			name = *s + strspn(*s, " \t\n");
			if (*name) {
				DB_TOKEN token;
				DB_REC_TARGET *t;
				token = lookup_db_target(y_db_main, name);
				if (DB_IS_NULL_TOKEN(token)) {
					fprintf(stderr, "%s: Illegal name '%s' in issrcfile\n", program,
						name);
					++errors;
				} else {
					t = read_db_target(y_db_main, token);
					t->flags.issrcfile = 1;
				}
			}
			++s;
		}
		free_saved_text(&saved_text);
	  }
	;

base_target
	: BASE_TARGET
	  LPAREN
	  { y_printf("__pp_select_list += (b_ "); }
	  text
	  RPAREN
	  { y_printf(")"); }
	;

update_if_changed
	: UPDATE_IF_CHANGED
	  LPAREN
	  { start_saving_text(); make_target = 0; }
	  local_target
	  optional_any_target
	  RPAREN
	  {
		char **s, *target1 = NULL, *target2 = NULL;
		for (s = saved_text; s && *s; ++s) {
			if (!isspace(**s)) {
				target1 = *s;
				++s;
				break;
			}
		}
		for (; s && *s; ++s) {
			if (!isspace(**s)) {
				target2 = *s;
				++s;
				break;
			}
		}
		fprintf(yyout, "%s: ", target1 + 1);
		if (target2) {
			fprintf(yyout, " %s\n", target2 + 1);
		} else {
			char *p = strrchr(target1, '/') + 1;
			fprintf(yyout, " %.*stmp_%s\n", (int) (p - (target1 + 1)),
				target1 + 1, p);
		}
		fprintf(yyout,
			"\t$(KBUILD_QUIET)cmp -s $< $@ || (($(check_writable);) && echo \"  Updating $@\" && cp -f $< $@)\n\n");
		fprintf(yyout, "# nodepend(%s)\n", target1);
		do_nodepend(target1);
		fprintf(yyout, "CLEAN\t+= %s\n", target1 + 1);
		free_saved_text(&saved_text);
	  }
	;

uses_asm_offsets
	: USES_ASM_OFFSETS
	  LPAREN
	    { start_saving_text(); }
	    text_list
	  RPAREN
	  {
		char **s, *t;
		stop_saving_text();
		if (!(y_feedback = fopen(feedback, "w+"))) {
			fprintf(stderr, "%s: Cannot create %s: %m\n", program,
				feedback);
			++errors;
			y_feedback = yyout;
		}
		for (s = saved_text; s && *s; ++s) {
			t = *s + strspn(*s, " \t\n");
			if (*t) {
				fprintf(y_feedback,
					"extra_aflags(%s $(obj_includelist /arch/$(ARCH)))\n",
					t);
				fprintf(y_feedback,
					"$(objfile %s):  $(objfile /arch/$(ARCH)/asm-offsets.h)\n",
					t);
			}
		}
		free_saved_text(&saved_text);
		if (y_feedback == yyout) {
			y_feedback = NULL;
		} else {
			/* y_feedback is read and closed in the lexer */
			fseek(y_feedback, 0, SEEK_SET);
		}
	  }
	;

hostcc
	: HOSTCC
	  LPAREN
	  { y_printf("__pp_select_list += (C_ "); make_target = 0; }
	  local_object_list
	  RPAREN
	  { y_printf(")"); }
	;

hostld
	: HOSTLD
	  LPAREN
	  { y_printf("__pp_select_list += (L_ "); make_target = 0; }
	  { start_saving_text(); }
	  local_target
	  {
		copy_saved_text(&hostld_binary);
		fprintf(yyout, "%s", *hostld_binary);
		stop_saving_text();
	  }
	  local_object_list
	  RPAREN
	  {
		y_printf(")\n");
		y_printf("CLEAN\t+= %s\n", *hostld_binary+1);
		free_saved_text(&hostld_binary);
	  }
	;

hostccld
	: HOSTCCLD
	  LPAREN
	    { start_saving_text(); }
	    text
	    { copy_saved_text(&hostccld_binary); }
	    { start_saving_text(); }
	    text_list
	    { copy_saved_text(&hostccld_objects); }
	  RPAREN
	  {
		char **text;
		stop_saving_text();
		if (!(y_feedback = fopen(feedback, "w+"))) {
			fprintf(stderr, "%s: Cannot create %s: %m\n", program,
				feedback);
			++errors;
			y_feedback = yyout;
		}
		fprintf(y_feedback, "hostcc(");
		for (text = hostccld_objects; *text; ++text) {
			if (!isspace(**text)) {
				fprintf(y_feedback, " %s", *text);
			}
		}
		fprintf(y_feedback, ")\n");
		fprintf(y_feedback, "hostld(%s", *hostccld_binary);
		for (text = hostccld_objects; *text; ++text) {
			if (!isspace(**text)) {
				fprintf(y_feedback, " %s", *text);
			}
		}
		fprintf(y_feedback, ")\n");
		free_saved_text(&hostccld_binary);
		free_saved_text(&hostccld_objects);
		if (y_feedback == yyout) {
			y_feedback = NULL;
		} else {
			/* y_feedback is read and closed in the lexer */
			fseek(y_feedback, 0, SEEK_SET);
		}
	  }
	;

setup
	: SETUP
	  LPAREN
	  { start_saving_text(); make_target = 0; }
	  local_target
	  RPAREN
	  {
		stop_saving_text();
		fprintf(yyout, "setup: %s\n", *saved_text+1);
		y_add_make_target("", *saved_text)->flags.setup = 1;
	  }
	;

installable
	: INSTALLABLE
	  LPAREN
	  { start_saving_text(); }
	  optional_config_list
	  {
	  	copy_saved_text(&config_list);
		start_saving_text();
		make_target = 1;
	  }
	  local_target_list
	  RPAREN
	  {
		char **s, *t;
		int configs = 0;
		stop_saving_text();
		if (!(y_feedback = fopen(feedback, "w+"))) {
			fprintf(y_feedback, "%s: Cannot create %s: %m\n", program,
				feedback);
			++errors;
			y_feedback = yyout;
		}
		fprintf(y_feedback, "create_objdir(");
		for (s = config_list; s && *s; ++s) {
			t = *s + strspn(*s, " \t\n");
			if (*t) {
				++configs;
				fprintf(y_feedback, "%s ", t);
			}
		}
		fprintf(y_feedback, ")\n");
		for (s = config_list; s && *s; ++s) {
			t = *s + strspn(*s, " \t\n");
			if (*t) {
				fprintf(y_feedback, "ifsel(%s)\n", t);
			}
		}
		/* saved_text contains targets */
		for (s = saved_text; s && *s; ++s) {
			t = *s + strspn(*s, " \t\n");
			if (*t) {
				fprintf(y_feedback, "installable: %s\n", t);
			}
		}
		while (configs--) {
			fprintf(y_feedback, "endif\n");
		}
		free_saved_text(&saved_text);
		free_saved_text(&config_list);
		if (y_feedback == yyout) {
			y_feedback = NULL;
		} else {
			/* y_feedback is read and closed in the lexer */
			fseek(y_feedback, 0, SEEK_SET);
		}
	  }
	;

optional_config_list
	: config_list
	| /* empty */
	;

config_list
	: config
	| config_list config
	;

config
	: CONFIG
	  { y_printf("%s", $1); ; free($1); }
	;

ifsel
	: IFSEL
	  LPAREN
	  CONFIG
	  { normal_config($3); do_ifsel($3, 0); free($3); }
	  RPAREN
	;

ifnsel
	: IFNSEL
	  LPAREN
	  CONFIG
	  RPAREN
	  { normal_config($3); y_printf("ifeq ($(subst n,,$(%s)),)", $3); free($3); }
	;

ifselmod
	: IFSELMOD
	  LPAREN
	  CONFIG
	  RPAREN
	  { normal_config($3); y_printf("ifeq ($(%s),m)", $3); free($3); }
	;

ifselnmod
	: IFSELNMOD
	  LPAREN
	  CONFIG
	  RPAREN
	  { normal_config($3); y_printf("ifeq ($(%s),y)", $3); free($3); }
	;

local_object_list
	: local_object
	| local_object_list local_object
	;

local_object
	: TEXT
	  {
		int l = strlen($1);
		if (strchr($1, '$')) {
			y_printf("$(patsubst %%,%s%%,%s)", l_curr_reldir + make_target,
				 $1);
		} else {
			if (l < 3 || $1[l - 2] != '.' || !strchr("oais", $1[l - 1])) {
				yyerror("not an object name");
			}
			if (strchr($1, '/')) {
				yyerror("local object names must not contain '/'");
			} else {
				y_printf("%s%s", l_curr_reldir + make_target, $1);
				y_add_make_target(l_curr_reldir, $1);
			}
		}
		free($1);
	  }
	;

local_target_list
	: local_target
	| local_target_list local_target
	;

local_target
	: TEXT
	  {
		int l = strlen($1);
		if (strchr($1, '$')) {
			y_printf("$(patsubst %%,%s%%,%s)", l_curr_reldir + make_target,
				 $1);
		} else {
			if (strchr($1, '/')) {
				yyerror("local target names must not contain '/'");
			} else {
				y_printf("%s%s", l_curr_reldir + make_target, $1);
				y_add_make_target(l_curr_reldir, $1);
			}
		}
		free($1);
	  }
	;

optional_any_target
	: any_target
	| /* empty */
	;

any_target_list
	: any_target
	| any_target_list any_target
	;

any_target
	: TEXT
	  {
		char name[100 + strlen(l_curr_reldir) + strlen($1)];
		if (*$1 == '/') {
			snprintf(name, sizeof (name), "%s", $1 + 1);
			y_add_make_target("", $1);
		} else if (strchr($1, '$')) {
			snprintf(name, sizeof (name),
				 "$(patsubst %%,%s%%,%s)", l_curr_reldir + make_target,
				 $1);
		} else {
			snprintf(name, sizeof (name), "%s%s",
				 l_curr_reldir + make_target, $1);
			standardize_name(name);
			y_add_make_target(l_curr_reldir, $1);
		}
		if (!(save_text(name))) {
			y_printf("%s", name);
		}
		free($1);
	  }
	;

link_subdir_list
	: link_subdir
	| link_subdir_list link_subdir
	;

link_subdir
	: TEXT
	  {
		if (*$1 == '/') {
#ifdef BRAIN_DEAD_CROSS_TREE_LINK_SUBDIRS
			y_normalize_path($1);
			y_printf("%s", $1);
#else
			yyerror("link_subdir directory names must not start with '/'");
#endif
		} else if (strstr($1, "../") || strstr($1, "/..")) {
#ifdef BRAIN_DEAD_CROSS_TREE_LINK_SUBDIRS
			char buf[strlen(l_curr_reldir) + strlen($1) + 1];
			sprintf(buf, "%s%s", l_curr_reldir, $1);
			y_normalize_path(buf);
			y_printf("%s", buf);
#else
			yyerror("link_subdir directory names must not contain '..'");
#endif
		} else if (strchr($1, '$')) {
			y_printf("$(patsubst %%,%s%%,%s)", l_curr_reldir, $1);
		} else {
			y_printf("%s%s", l_curr_reldir, $1);
		}
		free($1);
	  }
	;

optional_text_list
	: text_list
	| /* empty */
	;

text_list
	: text
	| text_list text
	;

text
	: TEXT
	  { y_printf("%s", $1); free($1); }
	;

%%

int
parse_makefile_in(PPDB * db)
{
	int ret;
	extern int yydebug;
	yydebug = 0;
	y_db_main = db;

	if (!(yyout = fopen(".tmp_global_makefile_in.2", "w"))) {
		fprintf(stderr,
			"%s: Cannot open .tmp_global_makefile_in.2 for writing, %m\n",
			program);
		return (1);
	}

	fputs
	    ("# Generated by linux/scripts/pp_makefile2 from Makefile.in files, do not edit.\n",
	     yyout);
	fputs
	    ("# This Makefile will not work without the variables that are exported from the\n",
	     yyout);
	fputs
	    ("# top level Makefile, do not try to use this Makefile directly.\n",
	     yyout);
	fputs("\n", yyout);
	fputs("MAKEFLAGS += -r\n", yyout);
	fputs(".SUFFIXES:\n", yyout);
	fputs("SUFFIXES:\n\n", yyout);
	fputs(".PHONY: .tmp_select\n", yyout);
	fputs(".tmp_select:\n", yyout);
	fputs("\t@/bin/true\n\n", yyout);
	fputs(".PHONY: .config\n", yyout);
	fputs(".config: ;\n", yyout);
	fputs("include .config\n\n", yyout);
	{
		const char *extra_text;
		if ((extra_text = getenv("extra_text"))) {
			/* Yes, this really is printf() with a variable format.  extra_text can
			 * contain newline and tab.
			 */
			y_printf(extra_text);
			fputs("\n\n", yyout);
		}
	}
	fputs("# Empty the selection file\n", yyout);
	fputs("__pp_makefile_dummy\t:= $(shell rm -f .tmp_select)\n", yyout);
	fputs("\n", yyout);

	ret = yyparse() || y_errors;
	fclose(yyout);
	return (ret);
}
